"""
Clients used to access third party systems.
"""

import logging
import os
from datetime import datetime, timedelta
from functools import wraps
from urllib.parse import parse_qs, urljoin, urlparse

import requests
from edx_rest_api_client.client import get_oauth_access_token

from enterprise_reporting.utils import retry_on_exception

LOGGER = logging.getLogger(__name__)


class EdxOAuth2APIMixin:
    """
    Base API Client Mixin for accessing edX IDA API endpoints.
    """
    @staticmethod
    def refresh_token(func):
        """
        Use this method decorator to ensure the JWT token is refreshed when needed.
        """
        @wraps(func)
        def inner(self, *args, **kwargs):
            """
            Before calling the wrapped function, we check if the JWT token is expired, and if so, re-connect.
            """
            if self.token_expired():
                self.connect()
            return func(self, *args, **kwargs)
        return inner


class EdxOAuth2APIClient(EdxOAuth2APIMixin):
    """
    Base API Client for accessing edX IDA API endpoints.
    """

    LMS_ROOT_URL = os.getenv('LMS_ROOT_URL', default='')
    ENTERPRISE_CATALOG_ROOT_URL = os.getenv('ENTERPRISE_CATALOG_ROOT_URL', default='https://enterprise-catalog.edx.org')
    LMS_OAUTH_HOST = os.getenv('LMS_OAUTH_HOST', default='')
    API_BASE_URL = LMS_ROOT_URL + '/api/'
    ACCESS_TOKEN_EXPIRY_THRESHOLD_IN_SECONDS = 60

    DEFAULT_VALUE_SAFEGUARD = object()

    def __init__(self, client_id=None, client_secret=None):
        """
        Connect to the REST API.
        """
        self.client_id = client_id or os.environ.get('LMS_OAUTH_KEY')
        self.client_secret = client_secret or os.environ.get('LMS_OAUTH_SECRET')
        self.expires_at = datetime.utcnow()
        self.access_token = None

    @retry_on_exception(max_retries=3, delay=2, backoff=2)
    def connect(self):
        """
        Connect to the REST API, authenticating with an access token retrieved with our client credentials.
        """
        url = urljoin(f'{self.LMS_OAUTH_HOST}/', 'oauth2/access_token')
        self.access_token, self.expires_at = get_oauth_access_token(url, self.client_id, self.client_secret)

    def token_expired(self):
        """
        Return True if the JWT token has expired, False if not.
        """
        return datetime.utcnow() > (self.expires_at - timedelta(seconds=self.ACCESS_TOKEN_EXPIRY_THRESHOLD_IN_SECONDS))

    @EdxOAuth2APIMixin.refresh_token
    def _requests(self, url, querystring):
        headers = {'Authorization': "JWT {}".format(self.access_token)}
        response = requests.get(
            url,
            headers=headers,
            params=querystring
        )
        response.raise_for_status()
        data = response.json()
        return data

    def _load_data(
            self,
            resource,
            detail_resource=None,
            resource_id=None,
            querystring=None,
            should_traverse_pagination=False,
            default=DEFAULT_VALUE_SAFEGUARD,
    ):
        """
        Loads a response from a call to one of the API endpoints.

        Arguments:
            resource: The endpoint resource name.
            detail_resource: The sub-resource to append to the path.
            resource_id: The resource ID for the specific detail to get from the endpoint.
            querystring: Optional query string parameters.
            should_traverse_pagination: Whether to traverse pagination or return paginated response.
            default: The default value to return in case of no response content.

        Returns
            (JSON): Data returned by the API.
        """
        default_val = default if default is not self.DEFAULT_VALUE_SAFEGUARD else {}
        querystring = querystring or {}
        path = resource
        if resource_id:
            path += '/' + resource_id
        if detail_resource:
            path += '/' + detail_resource

        url = urljoin(f'{self.API_BASE_URL}/', path)
        data = self._requests(url, querystring)

        if should_traverse_pagination:
            results = self.traverse_pagination(data, url)
            data = {
                'count': len(results),
                'next': None,
                'previous': None,
                'results': results,
            }

        return data or default_val

    def traverse_pagination(self, data, url):
        """
        Traverse a paginated API response.

        Extracts and concatenates "results" (list of dict) returned by DRF-powered
        APIs.

        Arguments:
            data (Dict): Current response dict from service API
            url (str): API endpoint path

        Returns:
            list of dict.
        """
        results = data.get('results', [])

        next_page = data.get('next')
        while next_page:
            querystring = parse_qs(urlparse(next_page).query, True)
            request_data = self._requests(url, querystring)

            results += request_data.get('results', [])
            next_page = request_data.get('next')

        return results



"""
Common query filters for all tables.
"""
from logging import getLogger
from typing import Optional, Tuple
from uuid import UUID

from enterprise_data.admin_analytics.database.filters.base import BaseFilter
from enterprise_data.admin_analytics.database.query_filters import BetweenQueryFilter, EqualQueryFilter, INQueryFilter
from enterprise_data.clients import EnterpriseApiClient
from enterprise_data.exceptions import EnterpriseApiClientException

LOGGER = getLogger(__name__)


class CommonFiltersMixin(BaseFilter):
    """
    Common filters.
    """
    @staticmethod
    def enterprise_customer_uuid_filter(enterprise_customer_uuid_params_key: str) -> EqualQueryFilter:
        """
        Filter by enterprise customer uuid.

        Arguments:
            enterprise_customer_uuid_params_key: The key against which value will be passed in the query.
        """
        return EqualQueryFilter(
            column='enterprise_customer_uuid',
            value_placeholder=enterprise_customer_uuid_params_key,
        )

    @staticmethod
    def date_range_filter(
        column: str,
        start_date_params_key: str,
        end_date_params_key: str
    ) -> BetweenQueryFilter:
        """
        Filter by a table column to be in the given date range.

        Arguments:
            column (str): The table column name to filter on.
            start_date_params_key (str): The start date key against which value will be passed in the query.
            end_date_params_key (str): The end date key against which value will be passed in the query.
        """
        return BetweenQueryFilter(
            column=column,
            range_placeholders=(start_date_params_key, end_date_params_key),
        )

    def enterprise_user_query_filter(  # pylint: disable=inconsistent-return-statements
        self,
        group_uuid: Optional[UUID],
        enterprise_customer_uuid: UUID
    ) -> Optional[Tuple[INQueryFilter, dict]]:
        """
        Get the query filter to filter enrollments for enterprise users in the given group.

        Arguments:
            group_uuid (UUID): The UUID of the group.
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.

        Returns:
            (INQueryFilter | None): The query filter to filter enrollments for enterprise users in the given group.
        """
        if not group_uuid:
            return None

        try:
            learners_in_group = EnterpriseApiClient.get_enterprise_user_ids_in_group(group_uuid)
        except EnterpriseApiClientException:
            LOGGER.exception(
                "Failed to get learners in group [%s] for enterprise [%s]",
                group_uuid,
                enterprise_customer_uuid,
            )
        else:
            params = {f'eu_{i}': i for i in learners_in_group}
            enterprise_user_id_in_filter = INQueryFilter(
                column='enterprise_user_id',
                values_placeholders=list(params.keys()),
            )
            return enterprise_user_id_in_filter, params



"""
Views for enterprise offers
"""
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import filters, viewsets

from enterprise_data.api.v1 import serializers
from enterprise_data.models import EnterpriseOffer

from .base import EnterpriseViewSetMixin


class EnterpriseOfferViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    Viewset for enterprise offers.
    """
    serializer_class = serializers.EnterpriseOfferSerializer
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend,)
    ordering_fields = '__all__'

    lookup_field = 'offer_id'

    filterset_fields = (
        'offer_id',
        'status'
    )

    def get_object(self):
        """
        This ensures that UUIDs with dashes are properly handled when requesting info about offers.

        Related to the work in EnterpriseOfferSerializer with `to_internal_value` and `to_representation`
        """
        self.kwargs['offer_id'] = self.kwargs['offer_id'].replace('-', '')
        return super().get_object()

    def get_queryset(self):
        enterprise_customer_uuid = self.kwargs['enterprise_id']
        return EnterpriseOffer.objects.filter(
            enterprise_customer_uuid=enterprise_customer_uuid,
        )



"""
Query filters for enrollments table.
"""
from enterprise_data.admin_analytics.database.filters.base import BaseFilter
from enterprise_data.admin_analytics.database.filters.mixins import CommonFiltersMixin
from enterprise_data.admin_analytics.database.query_filters import EqualQueryFilter


class FactCompletionAdminDashFilters(CommonFiltersMixin, BaseFilter):
    """
    Query filters for completions data in enrollments table.
    """

    @staticmethod
    def has_passed_filter() -> EqualQueryFilter:
        """
        Filter by has passed with fixed value 1.
        """
        return EqualQueryFilter(
            column='has_passed',
            value=1
        )



"""
Tests for leaderboard API.
"""

from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.tests.admin_analytics.mock_analytics_data import ENROLLMENTS, LEADERBOARD_RESPONSE
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
class TestLeaderboardAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsLeaderboardView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-leaderboard-list",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )
        get_enrollment_date_range_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_date_range',
            return_value=(datetime.now(), datetime.now())
        )

        get_enrollment_date_range_patcher.start()
        self.addCleanup(get_enrollment_date_range_patcher.stop)

    def verify_enrollment_data(self, results, results_count):
        """Verify the received enrollment data."""
        attrs = [
            "email",
            "course_title",
            "course_subject",
            "enroll_type",
            "enterprise_enrollment_date",
        ]

        assert len(results) == results_count

        filtered_data = []
        for enrollment in ENROLLMENTS:
            for result in results:
                if enrollment["email"] == result["email"]:
                    filtered_data.append({attr: enrollment[attr] for attr in attrs})
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch('enterprise_data.admin_analytics.database.tables.FactEngagementAdminDashTable.get_leaderboard_data_count')
    @patch('enterprise_data.admin_analytics.database.tables.FactEngagementAdminDashTable.get_all_leaderboard_data')
    def test_get(self, mock_get_all_leaderboard_data, mock_get_leaderboard_data_count):
        """
        Test the GET method for the AdvanceAnalyticsLeaderboardView works.
        """
        mock_get_all_leaderboard_data.return_value = LEADERBOARD_RESPONSE
        mock_get_leaderboard_data_count.return_value = len(LEADERBOARD_RESPONSE)
        response = self.client.get(self.url + '?page_size=2')
        assert response.status_code == status.HTTP_200_OK
        assert response['Content-Type'] == 'application/json'
        data = response.json()
        assert data['next'] == f'http://testserver{self.url}?page=2&page_size=2'
        assert data['previous'] is None
        assert data['current_page'] == 1
        assert data['num_pages'] == 6
        assert data['count'] == 12

        # fetch all records
        response = self.client.get(self.url + '?page_size=20')
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data['next'] is None
        assert data['previous'] is None
        assert data['current_page'] == 1
        assert data['num_pages'] == 1
        assert data['count'] == 12

    @patch('enterprise_data.admin_analytics.database.tables.FactEngagementAdminDashTable.get_leaderboard_data_count')
    @patch('enterprise_data.admin_analytics.database.tables.FactEngagementAdminDashTable.get_all_leaderboard_data')
    def test_get_csv(self, mock_get_all_leaderboard_data, mock_get_leaderboard_data_count):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView return correct CSV data.
        """
        mock_get_all_leaderboard_data.return_value = LEADERBOARD_RESPONSE[:5]
        mock_get_leaderboard_data_count.return_value = 5
        response = self.client.get(self.url, {'response_type': ResponseType.CSV.value})
        assert response.status_code == status.HTTP_200_OK

        # verify the response headers
        assert response['Content-Type'] == 'text/csv'

        content = b"".join(response.streaming_content).decode().splitlines()
        assert len(content) == 6

        # Verify CSV header.
        assert 'email,learning_time_hours,session_count,average_session_length,course_completion_count' == content[0]

        # verify the content
        assert 'paul77@example.org,4.4,,4.4,' in content
        assert 'seth57@example.org,2.7,,2.7,' in content
        assert 'weaverpatricia@example.net,2.6,,2.6,' in content
        assert 'webertodd@example.com,1.5,,1.5,' in content
        assert 'yferguson@example.net,1.3,,1.3,' in content



"""
Query filter for equal operation.
"""
from .base import QueryFilter


class EqualQueryFilter(QueryFilter):
    """
    Query filter for equal operation.
    """
    def __init__(self, column: str, value: str = None, value_placeholder: str = None):
        """
        Initialize the filter.

        This will also validate arguments.
        """
        self.validate_argument_exclusivity(value, value_placeholder)

        self.column = column
        self.value = value
        self.value_placeholder = value_placeholder

    def to_sql(self) -> str:
        if self.value is not None:
            return f'{self.column} = {self.value_to_sql(self.value)}'
        else:
            return f'{self.column} = %({self.value_placeholder})s'



"""
Test factories.
"""

from datetime import datetime, timedelta

import factory
from factory.faker import Faker
from faker import Factory as FakerFactory
from faker.providers import misc

from enterprise_data.models import EnterpriseAdminLearnerProgress, EnterpriseAdminSummarizeInsights

FAKER = FakerFactory.create()
FAKER.add_provider(misc)


class EnterpriseAdminLearnerProgressFactory(factory.django.DjangoModelFactory):
    """
    EnterpriseAdminLearnerProgress Factory.
    """
    class Meta:
        model = EnterpriseAdminLearnerProgress

    enterprise_customer_uuid = Faker('uuid4')
    enterprise_customer_name = FAKER.company()
    active_subscription_plan = True
    assigned_licenses = FAKER.pyint(min_value=0)
    activated_licenses = FAKER.pyint(min_value=0)
    assigned_licenses_percentage = FAKER.pyfloat(min_value=0.0, max_value=100)
    activated_licenses_percentage = FAKER.pyfloat(min_value=0.0, max_value=100)
    active_enrollments = FAKER.pyint(min_value=0)
    at_risk_enrollment_less_than_one_hour = FAKER.pyint(min_value=0)
    at_risk_enrollment_end_date_soon = FAKER.pyint(min_value=0)
    at_risk_enrollment_dormant = FAKER.pyint(min_value=0)
    created_at = datetime.now()


class EnterpriseAdminSummarizeInsightsFactory(factory.django.DjangoModelFactory):
    """
    EnterpriseAdminSummarizeInsights Factory.
    """
    class Meta:
        model = EnterpriseAdminSummarizeInsights

    enterprise_customer_uuid = Faker('uuid4')
    enterprise_customer_name = FAKER.company()
    enrolls = FAKER.pyint(min_value=0)
    enrolls_prior = FAKER.pyint(min_value=0)
    passed = FAKER.pyint(min_value=0)
    passed_prior = FAKER.pyint(min_value=0)
    engage = FAKER.pyint(min_value=0)
    engage_prior = FAKER.pyint(min_value=0)
    hours = FAKER.pyint(min_value=0)
    hours_prior = FAKER.pyint(min_value=0)
    contract_end_date = datetime.now() + timedelta(days=30)
    active_contract = True
    created_at = datetime.now()



"""
Module containing queries for the fact_enrollment_engagement_day_admin_dash table.
"""


class FactEngagementAdminDashQueries:
    """
    Queries related to the fact_enrollment_engagement_day_admin_dash table.
    """
    @staticmethod
    def get_learning_hours_and_daily_sessions_query():
        """
        Get the query to fetch the learning hours and daily sessions.
        """
        return """
            SELECT
                ROUND(SUM(learning_time_seconds) / 60 / 60, 1) as hours, SUM(is_engaged) as sessions
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                activity_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_engagement_count_query(query_filters):
        """
        Get the query to fetch the total number of engagements for an enterprise customer.
        """
        return f"""
            SELECT count(*)
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE {query_filters.to_sql()};
        """

    @staticmethod
    def get_all_engagement_query(query_filters):
        """
        Get the query to fetch all engagement data.
        """
        return f"""
            SELECT
                email, course_title, course_subject, enroll_type, activity_date,
                learning_time_seconds/3600 as learning_time_hours
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE {query_filters.to_sql()}
            ORDER BY activity_date DESC LIMIT %(limit)s OFFSET %(offset)s;
        """

    @staticmethod
    def get_top_courses_by_engagement_query(query_filters, record_count=10):
        """
        Get the query to fetch the learning time in hours by courses.

        Query should fetch the top N courses by user engagement. Where N is the value of record_count.

        Arguments:
            record_count (int): Number of records to fetch.

        Returns:
            (str): Query to fetch the learning time in hours by courses for the top courses by user engagement.
        """
        return f"""
            WITH filtered_data AS (
                SELECT
                    course_key,
                    course_title,
                    enroll_type,
                    (learning_time_seconds / 60.0 / 60.0) AS learning_time_hours,
                    activity_date
                FROM fact_enrollment_engagement_day_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_courses AS (
                SELECT
                    course_key,
                    SUM(learning_time_hours) as total_learning_time
                FROM filtered_data
                GROUP BY course_key
                ORDER BY total_learning_time DESC
                LIMIT {record_count}
            )
            SELECT
                d.course_key,
                d.course_title,
                d.enroll_type,
                ROUND(SUM(d.learning_time_hours)) AS learning_time_hours
            FROM filtered_data d
            JOIN top_10_courses tc
                ON d.course_key = tc.course_key
            GROUP BY d.course_key, d.course_title, d.enroll_type
            ORDER BY total_learning_time DESC;
        """

    @staticmethod
    def get_top_subjects_by_engagement_query(query_filters, record_count=10):
        """
        Get the query to fetch the learning time in hours by subjects.

        Query should fetch the top N subjects by user engagement. Where N is the value of record_count.

        Arguments:
            record_count (int): Number of records to fetch.

        Returns:
            (str): Query to fetch the learning time in hours by subjects for the top subjects by user engagement.
        """
        return f"""
            WITH filtered_data AS (
                SELECT
                    course_subject,
                    enroll_type,
                    (learning_time_seconds / 60.0 / 60.0) AS learning_time_hours,
                    activity_date
                FROM fact_enrollment_engagement_day_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_subjects AS (
                SELECT
                    course_subject,
                    SUM(learning_time_hours) as total_learning_time
                FROM filtered_data
                GROUP BY course_subject
                ORDER BY total_learning_time DESC
                LIMIT {record_count}
            )
            SELECT
                d.course_subject,
                d.enroll_type,
                ROUND(SUM(d.learning_time_hours)) AS learning_time_hours
            FROM filtered_data d
            JOIN top_10_subjects ts
                ON d.course_subject = ts.course_subject
            GROUP BY d.course_subject, d.enroll_type
            ORDER BY total_learning_time DESC;
        """

    @staticmethod
    def get_engagement_time_series_data_query(query_filters):
        """
        Get the query to fetch the completion time series data.

        Query should fetch the time series data with daily granularity.

        Returns:
            (str): Query to fetch the completion time series data.
        """
        return f"""
            SELECT activity_date, enroll_type, SUM(learning_time_seconds)/3600 as learning_time_hours
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE {query_filters.to_sql()}
            GROUP BY activity_date, enroll_type
            ORDER BY activity_date;
        """

    @staticmethod
    def get_engagement_data_for_leaderboard_query():
        """
        Get the query to fetch the engagement data for leaderboard.

        Query should fetch the engagement data for like learning time, session length of
        the enterprise learners to show in the leaderboard.

        Returns:
            (str): Query to fetch the engagement data for leaderboard.
        """
        return """
            SELECT
                email,
                ROUND(SUM(learning_time_seconds) / 3600, 1) as learning_time_hours,
                SUM(is_engaged) as session_count,
                CASE
                    WHEN SUM(is_engaged) = 0 THEN 0.0
                    ELSE ROUND(SUM(learning_time_seconds) / 3600 / SUM(is_engaged), 1)
                END AS average_session_length
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                (activity_date BETWEEN %(start_date)s AND %(end_date)s) AND
                is_engaged = 1
            GROUP BY email
            ORDER BY learning_time_hours DESC
            LIMIT %(limit)s OFFSET %(offset)s;
        """

    @staticmethod
    def get_completion_data_for_leaderboard_query(email_list: list):
        """
        Get the query to fetch the completions data for leaderboard.

        Query should fetch the completion data for like course completion count of
        the enterprise learners to show in the leaderboard.

        Arguments:
            email_list (str): List of emails to filter the completions data.

        Returns:
            (list<str>): Query to fetch the completions data for leaderboard.
        """
        return f"""
            SELECT email, count(course_key) as course_completion_count
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                (passed_date BETWEEN %(start_date)s AND %(end_date)s) AND
                has_passed = 1 AND
                email IN {str(tuple(email_list))}
            GROUP BY email;
        """

    @staticmethod
    def get_engagement_data_for_leaderboard_null_email_only_query():
        """
        Get the query to fetch the engagement data for leaderboard for NULL emails only.

        Query should fetch the engagement data for like learning time, session length of
        the enterprise learners to show in the leaderboard.

        Returns:
            (str): Query to fetch the engagement data for leaderboard.
        """
        return """
            SELECT
                email,
                ROUND(SUM(learning_time_seconds) / 3600, 1) as learning_time_hours,
                SUM(is_engaged) as session_count,
                CASE
                    WHEN SUM(is_engaged) = 0 THEN 0.0
                    ELSE ROUND(SUM(learning_time_seconds) / 3600 / SUM(is_engaged), 1)
                END AS average_session_length
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                (activity_date BETWEEN %(start_date)s AND %(end_date)s) AND
                is_engaged = 1 AND
                email is NULL
            GROUP BY email;
        """

    @staticmethod
    def get_completion_data_for_leaderboard_null_email_only_query():
        """
        Get the query to fetch the completions data for leaderboard for NULL emails.

        Query should fetch the completion data for like course completion count of
        the enterprise learners to show in the leaderboard.

        Returns:
            (list<str>): Query to fetch the completions data for leaderboard.
        """
        return """
            SELECT email, count(course_key) as course_completion_count
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                (passed_date BETWEEN %(start_date)s AND %(end_date)s) AND
                has_passed = 1 AND
                email is NULL
            GROUP BY email;
        """

    @staticmethod
    def get_leaderboard_data_count_query():
        """
        Get the query to fetch the leaderboard row count and null email counter.

        Query should fetch the count of rows for the leaderboard data for the enterprise customer.

        Returns:
            (str): Query to fetch the leaderboard row count.
        """
        return """
            SELECT
                COUNT(*) OVER () AS record_count
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                (activity_date BETWEEN %(start_date)s AND %(end_date)s) AND
                is_engaged = 1
            GROUP BY email
            LIMIT 1;
        """



"""
Tests for `./manage.py pre_warm_analytics_cache` management command.
"""
from datetime import datetime
from unittest import TestCase

from mock import MagicMock, patch
from pytest import mark

from django.core.management import call_command


@mark.django_db
class Test(TestCase):
    """
    Tests to validate the behavior of `./manage.py pre_warm_analytics_cache` management command.
    """
    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.enterprise_uuid = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'

        get_enrollment_date_range_patcher = patch(
            'enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_enrollment_date_range',
            return_value=(datetime.now(), datetime.now())
        )

        get_enrollment_date_range_patcher.start()
        self.addCleanup(get_enrollment_date_range_patcher.stop)

    @patch(
        'enterprise_data.admin_analytics.database.tables.fact_engagement_admin_dash.run_query',
        MagicMock(return_value=[])
    )
    @patch(
        'enterprise_data.admin_analytics.database.tables.fact_enrollment_admin_dash.run_query',
        MagicMock(return_value=[])
    )
    @patch(
        'enterprise_data.admin_analytics.database.tables.skills_daily_rollup_admin_dash.run_query',
        MagicMock(return_value=[])
    )
    @patch('enterprise_data.api.v1.views.analytics_enrollments.FactEnrollmentAdminDashTable.get_top_enterprises')
    @patch('enterprise_data.cache.decorators.cache.set')
    @patch('enterprise_data.cache.decorators.cache.get')
    def test_pre_warm_analytics_cache(self, mock_get_cache, mock_set_cache, mock_get_top_enterprises):
        """
        Validate that the command caches the analytics data for a large enterprise.
        """
        mock_get_top_enterprises.return_value = [
            self.enterprise_uuid
        ]
        mock_get_cache.return_value = MagicMock(is_found=False)

        call_command('pre_warm_analytics_cache')

        assert mock_get_cache.call_count == 23
        assert mock_set_cache.call_count == 23



"""
Query Filters for database tables.
"""

from .fact_completion_admin_dash import FactCompletionAdminDashFilters
from .fact_engagement_admin_dash import FactEngagementAdminDashFilters
from .fact_enrollment_admin_dash import FactEnrollmentAdminDashFilters



"""
Constants for enterprise data app.
"""

ANALYTICS_API_VERSION_ATTR = 'API_VERSION'
ANALYTICS_API_VERSION_0 = 'V0'
ANALYTICS_API_VERSION_1 = 'V1'

UUID4_REGEX = '[0-9a-f]{8}-?[0-9a-f]{4}-?4[0-9a-f]{3}-?[89ab][0-9a-f]{3}-?[0-9a-f]{12}'



"""
These settings are here to use during tests, because django requires them.

In a real-world use case, apps in this project are installed into other
Django applications, so these settings will not be used.
"""
import os
from os.path import abspath, dirname, join

from enterprise_data_roles.constants import (
    ENTERPRISE_DATA_ADMIN_ROLE,
    SYSTEM_ENTERPRISE_ADMIN_ROLE,
    SYSTEM_ENTERPRISE_OPERATOR_ROLE,
)


def here(*args):
    """
    Return the absolute path to a directory from this file.
    """
    return join(abspath(dirname(__file__)), *args)


def root(*args):
    """
    Return the absolute path to some file from the project's root.
    """
    return abspath(join(abspath(here('../..')), *args))


DATABASES = {
    "default": {
        "ENGINE": os.environ.get("DB_ENGINE", "django.db.backends.sqlite3"),
        "NAME": os.environ.get("DB_NAME", "default.db"),
        "USER": os.environ.get("DB_USER", ""),
        "PASSWORD": os.environ.get("DB_PASS", ""),
        "HOST": os.environ.get("DB_HOST", ""),
        "PORT": os.environ.get("DB_PORT", ""),
    },
}

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.admin",  # only used in DEBUG mode
    "django.contrib.messages",
    "django.contrib.staticfiles",

    "waffle",

    "enterprise_data",
    "enterprise_reporting",
    "enterprise_data_roles",
    "rules.apps.AutodiscoverRulesConfig",

    # drf-jwt
    "rest_framework_jwt",
)

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "crum.CurrentRequestUserMiddleware",
    "waffle.middleware.WaffleMiddleware",
]

AUTHENTICATION_BACKENDS = [
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend"
]

SESSION_ENGINE = "django.contrib.sessions.backends.file"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": {
                "django.contrib.auth.context_processors.auth",  # this is required for admin
                "django.contrib.messages.context_processors.messages",
            }
        }
    },
]

LMS_BASE_URL = "http://localhost:8000/"

BACKEND_SERVICE_EDX_OAUTH2_KEY = "analytics_api-backend-service-key"
BACKEND_SERVICE_EDX_OAUTH2_SECRET = "analytics_api-backend-service-secret"
BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL = "http://127.0.0.1:8000/oauth2"

ROOT_URLCONF = "enterprise_data.urls"

SECRET_KEY = "insecure-secret-key"

USE_TZ = True
TIME_ZONE = 'UTC'

SITE_NAME = 'analytics-data-api'

ENTERPRISE_REPORTING_DB_ALIAS = 'default'
ENROLLMENTS_PAGE_SIZE = 10000

# Required for use with edx-drf-extensions JWT functionality:
# USER_SETTINGS overrides for djangorestframework-jwt APISettings class
# See https://github.com/GetBlimp/django-rest-framework-jwt/blob/master/rest_framework_jwt/settings.py
JWT_AUTH = {
    'JWT_AUDIENCE': 'test-aud',
    'JWT_DECODE_HANDLER': 'edx_rest_framework_extensions.auth.jwt.decoder.jwt_decode_handler',
    'JWT_ISSUER': 'test-iss',
    'JWT_LEEWAY': 1,
    'JWT_SECRET_KEY': 'test-key',
    'JWT_SUPPORTED_VERSION': '1.0.0',
    'JWT_VERIFY_AUDIENCE': False,
    'JWT_VERIFY_EXPIRATION': True,
    'JWT_AUTH_HEADER_PREFIX': 'JWT',

    # JWT_ISSUERS enables token decoding for multiple issuers (Note: This is not a native DRF-JWT field)
    # We use it to allow different values for the 'ISSUER' field, but keep the same SECRET_KEY and
    # AUDIENCE values across all issuers.
    'JWT_ISSUERS': [
        {
            'ISSUER': 'test-issuer-1',
            'SECRET_KEY': 'test-secret-key',
            'AUDIENCE': 'test-audience',
        },
        {
            'ISSUER': 'test-issuer-2',
            'SECRET_KEY': 'test-secret-key',
            'AUDIENCE': 'test-audience',
        }
    ],
}

SYSTEM_TO_FEATURE_ROLE_MAPPING = {
    SYSTEM_ENTERPRISE_ADMIN_ROLE: [ENTERPRISE_DATA_ADMIN_ROLE],
    SYSTEM_ENTERPRISE_OPERATOR_ROLE: [ENTERPRISE_DATA_ADMIN_ROLE],
}



# Generated by Django 3.2.23 on 2024-02-12 14:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0038_enterpriseoffer_export_timestamp'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='enterpriselearnerenrollment',
            index=models.Index(fields=['enterprise_customer_uuid', 'enterprise_user_id', 'user_current_enrollment_mode'], name='enterprise__enterpr_6b0be8_idx'),
        ),
        migrations.AddIndex(
            model_name='enterpriselearnerenrollment',
            index=models.Index(fields=['enterprise_customer_uuid', 'offer_id', 'budget_id'], name='enterprise__enterpr_66e37f_idx'),
        ),
        migrations.AddIndex(
            model_name='enterpriselearnerenrollment',
            index=models.Index(fields=['enterprise_customer_uuid', 'user_current_enrollment_mode', 'coupon_code', 'offer_type'], name='enterprise__enterpr_1e8e98_idx'),
        ),
    ]



# pylint: skip-file

"""
Signals for enterprise-data.
"""


from edx_django_utils.cache import TieredCache

from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from enterprise_data.models import EnterpriseLearner, EnterpriseLearnerEnrollment
from enterprise_data.utils import get_cache_key


@receiver([post_save, post_delete], sender=[EnterpriseLearner, EnterpriseLearnerEnrollment])
def clear_cache(instance, **kwargs):
    """
    Signal receiver function to clear the cache of specified model.
    """
    cache_key = get_cache_key(
        resource='enterprise-learner',
        enterprise_customer=instance.enterprise_customer_uuid,
    )
    TieredCache.delete_all_tiers(cache_key)


# connecting signal with reciever.
post_save.connect(clear_cache, sender=EnterpriseLearner)
post_save.connect(clear_cache, sender=EnterpriseLearnerEnrollment)
post_delete.connect(clear_cache, sender=EnterpriseLearner)
post_delete.connect(clear_cache, sender=EnterpriseLearnerEnrollment)



"""
Classes that handle sending reports for enterprise customers with specific delivery methods.
"""

import logging
import os
from smtplib import SMTPException

import paramiko

from enterprise_reporting.constants import SFTP_OPS_GENIE_EMAIL_ALERT_EMAILS, SFTP_OPS_GENIE_EMAIL_ALERT_FROM_EMAIL
from enterprise_reporting.utils import (
    compress_and_encrypt,
    decrypt_string,
    retry_on_exception,
    send_email_with_attachment,
)

LOGGER = logging.getLogger(__name__)


class DeliveryMethod:
    """
    Base class for different types of Enterprise report delivery methods.
    """

    def __init__(self, reporting_config, password):
        """Initialize the SFTP Delivery Method."""
        self.enterprise_customer_name = reporting_config['enterprise_customer']['name']
        self.data_type = reporting_config['data_type']
        self.report_type = reporting_config['report_type']
        self.password = password
        self.encrypted_password = reporting_config['encrypted_password']
        self.pgp_encryption_key = reporting_config.get('pgp_encryption_key')
        self.enable_compression = reporting_config.get('enable_compression')

    def send(self, files):
        """Base method for sending files, to perform common sending logic."""
        if self.enable_compression:
            LOGGER.info(f'Encrypting data report for {self.enterprise_customer_name}')
            zip_password = decrypt_string(
                self.encrypted_password) if self.encrypted_password else self.encrypted_password
            zip_file = compress_and_encrypt(files, zip_password, self.pgp_encryption_key)
            return [zip_file]
        return [file.name for file in files]


class SMTPDeliveryMethod(DeliveryMethod):
    """
    Class that handles sending an enterprise report file via email.
    """

    REPORT_EMAIL_FROM_EMAIL = os.environ.get('SEND_EMAIL_FROM')
    REPORT_EMAIL_SUBJECT = '{enterprise_name} edX Learner Data'
    REPORT_EMAIL_BODY = """
    Please find the attached {type} data for courses on edX.
    For any questions or concerns, please contact your edX sales representative.
    Thanks,
    The edX for Business Team
    """

    def __init__(self, reporting_config, password):
        """Initialize the SMTP Delivery Method."""
        super().__init__(reporting_config, password)
        self._email = reporting_config['email']

    @property
    def email(self):
        return self._email

    @email.setter
    def email(self, value):
        """Make the internal email value a list if there are multiple emails."""
        self._email = value if isinstance(value, list) else [value]

    def send(self, files):
        """Send the given files through SMTP."""
        attachment_data = {file: None for file in super().send(files)}
        LOGGER.info(f'Emailing encrypted data to {self.enterprise_customer_name}')
        try:
            send_email_with_attachment(
                self.REPORT_EMAIL_SUBJECT.format(enterprise_name=self.enterprise_customer_name),
                self.REPORT_EMAIL_BODY.format(type=self.data_type),
                self.REPORT_EMAIL_FROM_EMAIL,
                self.email,
                attachment_data
            )
        except SMTPException:
            LOGGER.exception('Failed to send email report to {} for {}'.format(
                self.email,
                self.enterprise_customer_name
            ))
        else:
            LOGGER.info('Email report successfully sent to {} for {}'.format(
                self.email,
                self.enterprise_customer_name
            ))


class SFTPDeliveryMethod(DeliveryMethod):
    """
    Class that handles sending an enterprise report file via SFTP.
    """
    sender_email = SFTP_OPS_GENIE_EMAIL_ALERT_FROM_EMAIL
    receiver_emails = SFTP_OPS_GENIE_EMAIL_ALERT_EMAILS

    def __init__(self, reporting_config, password):
        """Initialize the SFTP Delivery Method."""
        super().__init__(reporting_config, password)
        self.hostname = reporting_config['sftp_hostname']
        self.port = reporting_config['sftp_port']
        self.username = reporting_config['sftp_username']
        self.file_path = reporting_config['sftp_file_path']

    @retry_on_exception(max_retries=3, delay=2, backoff=2)
    def send_over_sftp(self, data_reports):
        """
        Send the reports via SFTP, retry on exception.
        """
        LOGGER.info('Connecting via SFTP to remote host {} for {}'.format(
            self.hostname,
            self.enterprise_customer_name
        ))
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(
            hostname=self.hostname,
            port=self.port,
            username=self.username,
            password=self.password,
        )
        sftp = ssh.open_sftp()
        for report in data_reports:
            sftp.put(
                report,
                os.path.join(self.file_path, os.path.basename(report))
            )
        sftp.close()
        ssh.close()

    def send(self, files):
        """Send the given files through SFTP."""
        try:
            data_reports = super().send(files)
            self.send_over_sftp(data_reports)
        except Exception:  # pylint: disable=broad-except
            email_subject = f'SFTP transmission failed for {self.enterprise_customer_name}'
            email_body = f'Failed to send {self.data_type} report for {self.enterprise_customer_name}'
            LOGGER.exception(f'SFTP transmission failed for {self.enterprise_customer_name}')
        else:
            LOGGER.info(f'Successfully sent report via sftp for {self.enterprise_customer_name}')
            email_subject = f'SFTP transmission successful for {self.enterprise_customer_name}'
            email_body = f'SFTP transmission successful for {self.enterprise_customer_name}'

        send_email_with_attachment(
            subject=email_subject,
            body=email_body,
            from_email=self.sender_email,
            to_email=self.receiver_emails,
            attachment_data={},
        )



# Generated by Django 3.2.13 on 2022-06-15 19:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0027_enterpriselearnerenrollment_total_learning_time_seconds'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='offer_id',
            field=models.PositiveIntegerField(null=True),
        ),
    ]



"""
Module containing queries for the fact_enrollment_admin_dash table.
"""
from ..query_filters import QueryFilters


class FactEnrollmentAdminDashQueries:
    """
    Queries related to the fact_enrollment_admin_dash table.
    """
    @staticmethod
    def get_top_enterprises_query(count=10):
        """
        Get the query to fetch the top enterprises by enrollments.
        """
        return f"""
            SELECT enterprise_customer_uuid
            FROM fact_enrollment_admin_dash
            GROUP BY enterprise_customer_uuid
            ORDER BY COUNT(enterprise_customer_uuid) DESC LIMIT {count};
        """

    @staticmethod
    def get_enrollment_count_query(query_filters: QueryFilters) -> str:
        """
        Get the query to fetch the total number of enrollments for an enterprise customer.
        """
        return f"""
            SELECT count(*)
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()};
        """

    @staticmethod
    def get_all_enrollments_query(query_filters: QueryFilters) -> str:
        """
        Get the query to fetch all enrollments.
        """
        return f"""
            SELECT email, course_title, course_subject, enroll_type, enterprise_enrollment_date
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()}
            ORDER BY ENTERPRISE_ENROLLMENT_DATE DESC LIMIT %(limit)s OFFSET %(offset)s
        """

    @staticmethod
    def get_enrollment_date_range_query():
        """
        Get the query to fetch the enrollment date range.
        """
        return """
            SELECT
                MIN(enterprise_enrollment_date) AS min_enrollment_date,
                MAX(enterprise_enrollment_date) AS max_enrollment_date
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s;
        """

    @staticmethod
    def get_enrollment_and_course_count_query():
        """
        Get the query to fetch the enrollment and course count.
        """
        return """
            SELECT
                count(*) as enrolls, count(DISTINCT course_key) as courses
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_completion_count_query(query_filters: QueryFilters) -> str:
        """
        Get the query to fetch the completion count.
        """
        return f"""
            SELECT
                SUM(has_passed) as completions
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()};
        """

    @staticmethod
    def get_learning_hours_and_daily_sessions_query():
        """
        Get the query to fetch the learning hours and daily sessions.
        """
        return """
            SELECT
                ROUND(SUM(learning_time_seconds) / 60 / 60, 1) as hours, SUM(is_engaged) as sessions
            FROM fact_engagement_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                activity_date BETWEEN %(start_date)s AND %(end_date)s;
            """

    @staticmethod
    def get_top_courses_by_enrollments_query(query_filters: QueryFilters, record_count: int = 10) -> str:
        """
        Get the query to fetch the enrollment count by courses.

        Query will fetch the top N courses by enrollment count. Where N is the value of record_count.

        Arguments:
            query_filters (QueryFilters): List of query filters.
            record_count (int): Number of records to fetch.
        """
        return f"""
            WITH filtered_data AS (
                SELECT *
                FROM fact_enrollment_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_courses AS (
                SELECT course_key
                FROM filtered_data
                GROUP BY course_key
                ORDER BY COUNT(*) DESC
                LIMIT {record_count}
            )

            SELECT
                d.course_key,
                MAX(d.course_title) AS course_title,
                d.enroll_type,
                COUNT(*) AS enrollment_count
            FROM filtered_data d
            JOIN top_10_courses tc
                ON d.course_key = tc.course_key
            GROUP BY d.course_key, d.enroll_type;
        """

    @staticmethod
    def get_top_subjects_by_enrollments_query(query_filters: QueryFilters, record_count: int = 10) -> str:
        """
        Get the query to fetch the enrollment count by subjects.

        Query will fetch the top N subjects by enrollment count. Where N is the value of record_count.

        Arguments:
            record_count (int): Number of records to fetch.
            query_filters (QueryFilters): List of query filters.
        """
        return f"""
            WITH filtered_data AS (
                SELECT *
                FROM fact_enrollment_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_subjects AS (
                SELECT course_subject
                FROM filtered_data
                GROUP BY course_subject
                ORDER BY COUNT(*) DESC
                LIMIT {record_count}
            )
            SELECT
                d.course_subject,
                d.enroll_type,
                COUNT(*) AS enrollment_count
            FROM filtered_data d
            JOIN top_10_subjects ts
                ON d.course_subject = ts.course_subject
            GROUP BY d.course_subject, d.enroll_type;
        """

    @staticmethod
    def get_enrolment_time_series_data_query(query_filters: QueryFilters) -> str:
        """
        Get the query to fetch the enrollment time series data with daily aggregation.

        Arguments:
            query_filters (QueryFilters): A list of filters for this query.
        """
        return f"""
            SELECT enterprise_enrollment_date, enroll_type, COUNT(*) as enrollment_count
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()}
            GROUP BY enterprise_enrollment_date, enroll_type
            ORDER BY enterprise_enrollment_date;
        """

    @staticmethod
    def get_all_completions_query(
        query_filters: QueryFilters,
    ) -> str:
        """
        Get the query to fetch all completions.
        """
        return f"""
            SELECT email, course_title, course_subject, enroll_type, passed_date
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()}
            ORDER BY passed_date DESC LIMIT %(limit)s OFFSET %(offset)s
        """

    @staticmethod
    def get_top_courses_by_completions_query(query_filters: QueryFilters, record_count=10) -> str:
        """
        Get the query to fetch the completion count by courses.

        Query should fetch the top N courses by completion count. Where N is the value of record_count.

        Arguments:
            query_filters (QueryFilters): List of query filters.
            record_count (int): Number of records to fetch.

        Returns:
            (str): Query to fetch the enrollment count by courses for the top courses by enrollment count.
        """
        return f"""
            WITH filtered_data AS (
                SELECT
                    course_key,
                    course_title,
                    enroll_type,
                    passed_date
                FROM fact_enrollment_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_courses AS (
                SELECT
                    course_key,
                    COUNT(*) AS total_completion_count
                FROM filtered_data
                GROUP BY course_key
                ORDER BY total_completion_count DESC
                LIMIT {record_count}
            )
            SELECT
                d.course_key,
                d.course_title,
                d.enroll_type,
                COUNT(*) AS completion_count
            FROM filtered_data d
            JOIN top_10_courses tc
                ON d.course_key = tc.course_key
            GROUP BY d.course_key, d.course_title, d.enroll_type
            ORDER BY total_completion_count DESC;
        """

    @staticmethod
    def get_top_subjects_by_completions_query(query_filters: QueryFilters, record_count=10) -> str:
        """
        Get the query to fetch the completion count by subjects.

        Query should fetch the top N subjects by completion count. Where N is the value of record_count.

        Arguments:
            query_filters (QueryFilters): List of query filters.
            record_count (int): Number of records to fetch.

        Returns:
            (str): Query to fetch the completion count by subjects for the top subjects by completion count.
        """
        return f"""
            WITH filtered_data AS (
                SELECT
                    course_subject,
                    enroll_type,
                    passed_date
                FROM fact_enrollment_admin_dash
                WHERE {query_filters.to_sql()}
            ),
            top_10_subjects AS (
                SELECT
                    course_subject,
                    COUNT(*) AS total_completion_count
                FROM filtered_data
                GROUP BY course_subject
                ORDER BY total_completion_count DESC
                LIMIT {record_count}
            )
            SELECT
                d.course_subject,
                d.enroll_type,
                COUNT(*) AS completion_count
            FROM filtered_data d
            JOIN top_10_subjects ts
                ON d.course_subject = ts.course_subject
            GROUP BY d.course_subject, d.enroll_type
            ORDER BY total_completion_count DESC;
        """

    @staticmethod
    def get_completions_time_series_data_query(
        query_filters: QueryFilters,
    ) -> str:
        """
        Get the query to fetch the completion time series data.

        Query should fetch the time series data with daily granularity.

        Returns:
            (str): Query to fetch the completion time series data.
        """
        return f"""
            SELECT passed_date, enroll_type, count(course_key) as completion_count
            FROM fact_enrollment_admin_dash
            WHERE {query_filters.to_sql()}
            GROUP BY passed_date, enroll_type
            ORDER BY passed_date;
        """



"""
Base views for enterprise data api v1.
"""
import math

from edx_rbac.mixins import PermissionRequiredMixin
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework.exceptions import NotFound
from rest_framework.response import Response
from rest_framework.utils.urls import remove_query_param, replace_query_param

from enterprise_data.constants import ANALYTICS_API_VERSION_1


class EnterpriseViewSetMixin(PermissionRequiredMixin):
    """
    Base class for all Enterprise view sets.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = DefaultPagination
    permission_required = 'can_access_enterprise'
    API_VERSION = ANALYTICS_API_VERSION_1

    def paginate_queryset(self, queryset):
        """
        Allows no_page query param to skip pagination
        """
        if 'no_page' in self.request.query_params:
            return None
        return super().paginate_queryset(queryset)


class AnalyticsPaginationMixin:
    """
    Mixin that provides utility methods to allow pagination on views.
    """
    page_query_param = 'page'
    page_size_query_param = 'page_size'

    def get_next_link(self, request, page_number, page_count):
        """
        Get the link to the next page.

        Arguments:
            request (Request): The request object.
            page_number (int): The current page number.
            page_count (int): The total number of pages.

        Returns:
            str: The link to the next page.
        """
        if page_number >= page_count:
            return None
        return replace_query_param(
            url=request.build_absolute_uri(),
            key=self.page_query_param,
            val=page_number + 1,
        )

    def get_previous_link(self, request, page_number):
        """
        Get the link to the previous page.

        Arguments:
            request (Request): The request object.
            page_number (int): The current page number.

        Returns:
            str: The link to the previous page.
        """
        if page_number <= 1:
            return None
        url = request.build_absolute_uri()
        next_page = page_number - 1
        if next_page == 1:
            return remove_query_param(url, self.page_query_param)
        return replace_query_param(url, self.page_query_param, next_page)

    def get_paginated_response(self, request, records, page, page_size, total_count):
        """
        Get pagination data.

        Arguments:
            request (Request): The request object.
            records (list): The records to return.
            page (int): The current page number.
            page_size (int): The number of records per page.
            total_count (int): The total number of records

        Returns:
            (Response): The pagination data.
        """
        page_count = math.ceil(total_count / page_size)
        if page_count > 0 and (page <= 0 or page > page_count):
            raise NotFound('Invalid page.')

        return Response({
            'next': self.get_next_link(request, page, page_count),
            'previous': self.get_previous_link(request, page),
            'count': total_count,
            'num_pages': page_count,
            'current_page': page,
            'results': records,
        })



"""
Custom Exceptions related to enterprise data.
"""


class EnterpriseDataException(Exception):
    """
    Base exception for enterprise data.
    """

    def __init__(self, message: str):
        """
        Initialize the exception with a message.
        """
        super().__init__(message)


class EnterpriseApiClientException(EnterpriseDataException):
    """
    Exception for enterprise API client.
    """



"""
Constants for enterprise_reporting.
"""


SFTP_OPS_GENIE_EMAIL_ALERT_FROM_EMAIL = "enterprise-integrations@edx.org"
SFTP_OPS_GENIE_EMAIL_ALERT_EMAILS = ['enterprise-reporting-sftp@2u-internal.opsgenie.net']



"""
Utility functions for interacting with the database.
"""
from contextlib import closing
from logging import getLogger

from mysql.connector import connect

from django.conf import settings

from enterprise_data.utils import timeit

LOGGER = getLogger(__name__)


def get_db_connection(database=settings.ENTERPRISE_REPORTING_DB_ALIAS):
    """
    Get a connection to the database.

    Returns:
        (mysql.connector.connection.MySQLConnection): The database connection.
    """
    return connect(
        host=settings.DATABASES[database]['HOST'],
        port=settings.DATABASES[database]['PORT'],
        database=settings.DATABASES[database]['NAME'],
        user=settings.DATABASES[database]['USER'],
        password=settings.DATABASES[database]['PASSWORD'],
    )


@timeit
def run_query(query, params: dict = None, as_dict=False):
    """
    Run a query on the database and return the results.

    Arguments:
        query (str): The query to run.
        params (dict): The parameters to pass to the query.
        as_dict (bool): When True, return the results as a dictionary.

    Returns:
        (list | dict): The results of the query.
    """
    try:
        with closing(get_db_connection()) as connection:
            with closing(connection.cursor()) as cursor:
                cursor.execute(query, params=params)
                if as_dict:
                    columns = [column[0] for column in cursor.description]
                    return [dict(zip(columns, row)) for row in cursor.fetchall()]
                else:
                    return cursor.fetchall()
    except Exception:
        LOGGER.exception(f'[run_query]: run_query failed for query "{query}".')
        raise



"""
Query filter for between operation.
"""
from .base import QueryFilter


class BetweenQueryFilter(QueryFilter):
    """
    Query filter for between operation.
    """

    def __init__(self, column: str, _range: tuple = None, range_placeholders: tuple = None):
        """
        Initialize the filter.

        This will also validate arguments.
        """
        self.validate_argument_exclusivity(_range, range_placeholders)

        self.column = column
        self.range = _range
        self.range_placeholders = range_placeholders

    def to_sql(self) -> str:
        if self.range:
            lower, upper = self.range
            return f'{self.column} BETWEEN {self.value_to_sql(lower)} AND {self.value_to_sql(upper)}'
        else:
            lower_placeholder, upper_placeholder = self.range_placeholders
            return f'{self.column} BETWEEN %({lower_placeholder})s AND %({upper_placeholder})s'



"""
Query filters for enrollments table.
"""
from enterprise_data.admin_analytics.database.filters.base import BaseFilter
from enterprise_data.admin_analytics.database.filters.mixins import CommonFiltersMixin


class FactEnrollmentAdminDashFilters(CommonFiltersMixin, BaseFilter):
    """
    Query filters for enrollments table.
    """



"""
Query filters for engagments data.
"""
from enterprise_data.admin_analytics.database.filters.base import BaseFilter
from enterprise_data.admin_analytics.database.filters.mixins import CommonFiltersMixin


class FactEngagementAdminDashFilters(CommonFiltersMixin, BaseFilter):
    """
    Query filters for engagments data.
    """



"""
Utility functions for fetching data from the database.
"""
from logging import getLogger

import pandas

from enterprise_data.admin_analytics.database import run_query

LOGGER = getLogger(__name__)


def fetch_max_enrollment_datetime():
    """
    Fetch the latest created date from the enterprise_learner_enrollment table.

    created will be same for all records as this is added at the time of data load. Which is when the async process
    populates the data in the table. We can use this to get the latest data load time.
    """
    query = "SELECT MAX(created) FROM enterprise_learner_enrollment"
    results = run_query(query)
    if not results:
        return None
    return pandas.to_datetime(results[0][0], utc=True)



"""
Views for fetching leaderboard data.
"""

from datetime import date
from logging import getLogger

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.viewsets import ViewSet

from django.http import StreamingHttpResponse

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.admin_analytics.database.tables import FactEngagementAdminDashTable, FactEnrollmentAdminDashTable
from enterprise_data.api.v1.serializers import AdvanceAnalyticsQueryParamSerializer
from enterprise_data.api.v1.views.base import AnalyticsPaginationMixin
from enterprise_data.renderers import LeaderboardCSVRenderer

LOGGER = getLogger(__name__)


class AdvanceAnalyticsLeaderboardView(AnalyticsPaginationMixin, ViewSet):
    """
    View to handle requests for enterprise leaderboard data.

    Here is the list of URLs that are handled by this view:
    1. `enterprise_data_api_v1.enterprise-admin-analytics-leaderboard-list`: Get leaderboard data.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ('get', )

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def list(self, request, enterprise_uuid):
        """
        Get individual leaderboard data for the enterprise.
        """
        # Remove hyphens from the UUID
        enterprise_uuid = enterprise_uuid.replace('-', '')

        serializer = AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)
        min_enrollment_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_uuid,
        )

        # get values from query params or use default values
        start_date = serializer.data.get('start_date', min_enrollment_date)
        end_date = serializer.data.get('end_date', date.today())
        page = serializer.data.get('page', 1)
        page_size = serializer.data.get('page_size', 100)
        total_count = FactEngagementAdminDashTable().get_leaderboard_data_count(
            enterprise_customer_uuid=enterprise_uuid,
            start_date=start_date,
            end_date=end_date,
        )
        leaderboard = FactEngagementAdminDashTable().get_all_leaderboard_data(
            enterprise_customer_uuid=enterprise_uuid,
            start_date=start_date,
            end_date=end_date,
            limit=page_size,
            offset=(page - 1) * page_size,
            total_count=total_count,
        )
        response_type = request.query_params.get('response_type', ResponseType.JSON.value)

        LOGGER.info(
            'Leaderboard data requested for enterprise [%s] from [%s] to [%s]',
            enterprise_uuid,
            start_date,
            end_date,
        )

        if response_type == ResponseType.CSV.value:
            filename = f'Leaderboard, {start_date} - {end_date}.csv'

            return StreamingHttpResponse(
                LeaderboardCSVRenderer().render(self._stream_serialized_data(
                    enterprise_uuid, start_date, end_date, total_count
                )),
                content_type='text/csv',
                headers={'Content-Disposition': f'attachment; filename="{filename}"'},
            )

        return self.get_paginated_response(
            request=request,
            records=leaderboard,
            page=page,
            page_size=page_size,
            total_count=total_count,
        )

    @staticmethod
    def _stream_serialized_data(enterprise_uuid, start_date, end_date, total_count, page_size=50000):
        """
        Stream the serialized data.
        """
        offset = 0
        while offset < total_count:
            leaderboard = FactEngagementAdminDashTable().get_all_leaderboard_data(
                enterprise_customer_uuid=enterprise_uuid,
                start_date=start_date,
                end_date=end_date,
                limit=page_size,
                offset=offset,
                total_count=total_count,
            )
            yield from leaderboard
            offset += page_size



"""
Mixins for enterprise data tests.
"""


from edx_rest_framework_extensions.auth.jwt.cookies import jwt_cookie_name
from edx_rest_framework_extensions.auth.jwt.tests.utils import generate_jwt_token, generate_unversioned_payload


class JWTTestMixin:
    """
    Mixin for JWT test related utils.
    """

    def set_jwt_cookie(self, system_wide_role='enterprise_admin', context='some_context'):
        """
        Set jwt token in cookies
        """
        role_data = f'{system_wide_role}'
        if context is not None:
            role_data += f':{context}'

        payload = generate_unversioned_payload(self.user)
        payload.update({
            'roles': [role_data]
        })
        jwt_token = generate_jwt_token(payload)

        self.client.cookies[jwt_cookie_name()] = jwt_token



"""
Base class to store the table information.
"""


class BaseTable:
    """
    Singleton class to store the table information.
    """
    instance = None

    def __new__(cls):
        """
        Singleton method to create a single instance of the table.
        """
        if cls.instance is None:
            cls.instance = super(BaseTable, cls).__new__(cls)
        return cls.instance



"""
Test delivery methods.
"""

import unittest

import ddt
from mock import MagicMock, patch

from enterprise_reporting.delivery_method import SFTPDeliveryMethod, SMTPDeliveryMethod
from enterprise_reporting.utils import encrypt_string

from .utils import create_files, verify_compressed


@ddt.ddt
class TestDeliveryMethod(unittest.TestCase):
    """
    Tests about delivery methods.
    """
    def setUp(self):
        self.encrypted_password = 'alohomora'
        self.password = 'magic_is_might'
        self.reporting_config = {
            'enterprise_customer': {
                'uuid': "abc",
                'name': 'bleh-bleh'
            },
            'data_type': 'progress',
            'report_type': 'csv',
            'encrypted_password': self.encrypted_password,
            'email': 'harry@gryffindor.hogwarts',
            'sftp_hostname': 'hogwarts_express',
            'sftp_port': 4444,
            'sftp_username': 'harry_potter',
            'sftp_file_path': 'platform/3/4'
        }

    @ddt.data(
        (SMTPDeliveryMethod, True),
        (SMTPDeliveryMethod, False),
        (SFTPDeliveryMethod, True),
        (SFTPDeliveryMethod, False),
    )
    @ddt.unpack
    def test_delivery_method_zipfile_password(self, delivery_method_class, enable_compression):
        """
        Test that `encrypted_password` is applied to zipfile irrespective of the delivery method.
        """
        self.reporting_config['encrypted_password'] = encrypt_string(self.encrypted_password)
        self.reporting_config['enable_compression'] = enable_compression
        delivery_method = delivery_method_class(self.reporting_config, self.password)
        file_data = [
            {
                'name': 'A History of Magic.txt',
                'size': 1000
            },
            {
                'name': 'Quidditch Through the Ages.txt',
                'size': 500
            },
            {
                'name': 'Quidditch Through the Ages.txt',
                'size': 500
            },
        ]
        files, total_original_size = create_files(file_data)
        delivery_files = super(delivery_method_class, delivery_method).send([file['file'] for file in files])
        if enable_compression:
            assert len(delivery_files) == 1  # there should be only one compressed file.
            verify_compressed(self, delivery_files[0], files, total_original_size, self.encrypted_password)
        else:
            assert len(delivery_files) == len(delivery_files)
            assert delivery_files == [file['file'].name for file in files]

    @patch('enterprise_reporting.delivery_method.send_email_with_attachment')
    def test_verify_sftp_exception_handling(self, mock_send_email_with_attachment):
        """
        Verify that SFTP transmission related exceptions are being handled correctly.
        """
        file_data = [
            {
                'name': 'A History of Magic.txt',
                'size': 1000
            },
            {
                'name': 'Quidditch Through the Ages.txt',
                'size': 500
            },
            {
                'name': 'Quidditch Through the Ages.txt',
                'size': 500
            },
        ]
        files, total_original_size = create_files(file_data)
        sftp_delivery_method = SFTPDeliveryMethod(self.reporting_config, self.password)

        # Verify failed SFTP transmission.
        with patch('paramiko.SSHClient.connect', side_effect=Exception('SFTP transmission failed')):
            sftp_delivery_method.send([file['file'] for file in files])
            mock_send_email_with_attachment.assert_called_with(
                subject='SFTP transmission failed for bleh-bleh',
                body='Failed to send progress report for bleh-bleh',
                from_email='enterprise-integrations@edx.org',
                to_email=['enterprise-reporting-sftp@2u-internal.opsgenie.net'],
                attachment_data={},
            )

        # Verify successful SFTP transmission.
        with patch('paramiko.SSHClient', MagicMock()):
            sftp_delivery_method.send([file['file'] for file in files])
            mock_send_email_with_attachment.assert_called_with(
                subject='SFTP transmission successful for bleh-bleh',
                body='SFTP transmission successful for bleh-bleh',
                from_email='enterprise-integrations@edx.org',
                to_email=['enterprise-reporting-sftp@2u-internal.opsgenie.net'],
                attachment_data={},
            )



"""
Test the utility functions in the admin_analytics app for data loading operations.
"""
from mock import patch

from django.test import TestCase

from enterprise_data.admin_analytics.data_loaders import fetch_max_enrollment_datetime


class TestDataLoaders(TestCase):
    """
    Test suite for the utility functions in the admin_analytics package for data loading operations.
    """

    def test_fetch_max_enrollment_datetime(self):
        """
        Validate the fetch_max_enrollment_datetime function.
        """
        with patch('enterprise_data.admin_analytics.data_loaders.run_query') as mock_run_query:
            mock_run_query.return_value = [['2024-07-26']]

            max_enrollment_date = fetch_max_enrollment_datetime()
            self.assertEqual(max_enrollment_date.strftime('%Y-%m-%d'), '2024-07-26')

            # Validate the case where the query returns an empty result.
            mock_run_query.return_value = []
            max_enrollment_date = fetch_max_enrollment_datetime()
            self.assertIsNone(max_enrollment_date)



"""
Client for connecting to the LMS Enterprise endpoints.
"""


import logging
import os
from collections import OrderedDict
from urllib.parse import urljoin

from requests.exceptions import ConnectionError, Timeout  # pylint: disable=redefined-builtin

from enterprise_reporting import utils
from enterprise_reporting.clients import EdxOAuth2APIClient

LOGGER = logging.getLogger(__name__)


class EnterpriseCatalogAPIClient(EdxOAuth2APIClient):
    """
    Client for connect to the Enterprise Catalog Service API.
    """
    API_BASE_URL = EdxOAuth2APIClient.ENTERPRISE_CATALOG_ROOT_URL + '/api/v1/'

    ENTERPRISE_CATALOGS_ENDPOINT = 'enterprise-catalogs/'
    GET_CONTENT_METADATA_ENDPOINT = 'enterprise-catalogs/{}/get_content_metadata'

    ENTERPRISE_REPORTING_ENDPOINT = 'enterprise_customer_reporting'

    PAGE_SIZE = os.getenv('PAGE_SIZE', default=1000)

    @staticmethod
    def _get_formatted_subjects(item):
        """
        Returns formatted subject for course and programs
        """
        subjects = item.get('subjects', [])
        formatted_subjects = []
        for subject in subjects:
            try:
                if isinstance(subject, str):
                    formatted_subjects.append(subject)
                elif isinstance(subject, dict):
                    subject_name = subject.get('name')
                    if subject_name:
                        formatted_subjects.append(subject_name)
                else:
                    error_msg = "Subject is not a string or a dictionary: {}".format(subject)
                    LOGGER.error("[Enterprise Reporting]. %s", error_msg)
                    raise TypeError(error_msg)
            except AttributeError as error:
                LOGGER.exception("[Enterprise Reporting]. Item: %s", item)
                raise error
        return formatted_subjects

    def transform_get_content_metadata(self, traversed_metadata):
        """
        Helper method to transform a response (already traversed pagination) from the enterprise-catalog service's `get_content_metadata`.
        endpoint.

        Note: as of June 2022 the data source for content metadata changed from edx platform to the enterprise catalog
        service. As such, content metadata coming from the enterprise catalog service needs to be formatted to match
        the old endpoints data structure.
        """
        content_metadata = OrderedDict()
        for item in traversed_metadata:
            content_id = utils.get_content_metadata_item_id(item)

            # Check if the item is a courserun
            if item.get('content_type') == 'courserun':
                # Courserun content metadata needs no massaging as it's the same in platform as it is in the
                # enterprise catalog service
                formatted_metadata = item
            else:
                # Course content metadata differs slightly between platform and the catalog service, so we need to
                # massage the response ie filter out new, not expected fields
                item_crs = item.get('course_runs', [])
                formatted_course_runs = []
                for cr in item_crs:
                    # While courseruns as a content metadata item did not change between old and new api endpoints,
                    # the courserun attribute of a course did, so we need to do some formatting.
                    formatted_course_run = {
                        'key': cr.get('key'),
                        'enrollment_start': cr.get('enrollment_start'),
                        'enrollment_end': cr.get('enrollment_end'),
                        'go_live_date': cr.get('go_live_date'),
                        'start': cr.get('start'),
                        'end': cr.get('end'),
                        'modified': cr.get('modified'),
                        'availability': cr.get('availability'),
                        'status': cr.get('status'),
                        'pacing_type': cr.get('pacing_type'),
                        # New endpoint uses `type` instead of enrollment mode
                        'enrollment_mode': cr.get('type'),
                        'min_effort': cr.get('min_effort'),
                        'max_effort': cr.get('max_effort'),
                        'weeks_to_complete': cr.get('weeks_to_complete'),
                        'estimated_hours': cr.get('estimated_hours'),
                        'first_enrollable_paid_seat_price': cr.get('first_enrollable_paid_seat_price'),
                        'is_enrollable': cr.get('is_enrollable'),
                    }
                    formatted_course_runs.append(formatted_course_run)

                formatted_subjects = self._get_formatted_subjects(item)
                formatted_metadata = {
                    'active': item.get('active'),
                    'aggregation_key': item.get('aggregation_key'),
                    'card_image_url': item.get('card_image_url'),
                    'content_type': item.get('content_type'),
                    'course_ends': item.get('course_ends'),
                    'course_runs': formatted_course_runs,
                    'end_date': item.get('end_date'),
                    'enrollment_url': item.get('enrollment_url'),
                    'full_description': item.get('full_description'),
                    'image_url': item.get('image_url'),
                    'key': item.get('key'),
                    'languages': item.get('languages'),
                    'organizations': item.get('organizations'),
                    'seat_types': item.get('seat_types'),
                    'short_description': item.get('short_description'),
                    'skill_names': item.get('skill_names'),
                    'skills': item.get('skills'),
                    'subjects': formatted_subjects,
                    'title': item.get('title'),
                    'uuid': item.get('uuid'),
                }

            content_metadata[content_id] = formatted_metadata

        return content_metadata

    def get_content_metadata(self, enterprise_customer_catalogs):
        """Return all content metadata contained in the catalogs associated with a reporting config."""
        content_metadata = OrderedDict()
        for catalog in enterprise_customer_catalogs.get('results', []):
            traversed_metadata = self._load_data(
                self.GET_CONTENT_METADATA_ENDPOINT.format(catalog['uuid']),
                should_traverse_pagination=True,
                querystring={'page_size': self.PAGE_SIZE},
            )
            transformed_metadata = self.transform_get_content_metadata(traversed_metadata.get('results'))
            content_metadata.update(transformed_metadata)

        # We only made this a dictionary to help filter out duplicates by a common key. We just want values now.
        return list(content_metadata.values())

    def get_customer_catalogs(self, enterprise_customer_uuid):
        """Return all catalog uuids owned by an Enterprise Customer."""
        return self._load_data(
            self.ENTERPRISE_CATALOGS_ENDPOINT,
            should_traverse_pagination=True,
            querystring={
                'enterprise_customer': enterprise_customer_uuid,
                'page_size': self.PAGE_SIZE,
            },
        )


class EnterpriseAPIClient(EdxOAuth2APIClient):
    """
    Client for connecting to the Enterprise API.
    """
    API_BASE_URL = urljoin(EdxOAuth2APIClient.LMS_ROOT_URL + '/', 'enterprise/api/v1/')

    ENTERPRISE_REPORTING_ENDPOINT = 'enterprise_customer_reporting'

    PAGE_SIZE = os.getenv('PAGE_SIZE', default=1000)

    def get_all_enterprise_reporting_configs(self, **kwargs):
        """
        Query the enterprise customer reporting endpoint for all available configs.
        """
        return self._load_data(
            self.ENTERPRISE_REPORTING_ENDPOINT,
            should_traverse_pagination=True,
            **kwargs
        )

    def get_enterprise_reporting_configs(self, enterprise_customer_uuid, **kwargs):
        """
        Query the enterprise customer reporting endpoint for a given enterprise customer.
        """
        return self._load_data(
            self.ENTERPRISE_REPORTING_ENDPOINT,
            querystring={'enterprise_customer': enterprise_customer_uuid},
            should_traverse_pagination=True,
            **kwargs
        )


class EnterpriseDataApiClient(EdxOAuth2APIClient):
    """
    Client for connecting to the Enterprise Data API.
    """

    API_BASE_URL = urljoin(os.getenv('ANALYTICS_API_URL', default='') + '/', 'enterprise/api/v0')
    PAGE_SIZE = os.getenv('PAGE_SIZE', default=1000)

    def get_enterprise_enrollments(self, enterprise_customer_uuid):
        return self._load_data(
            'enterprise',
            resource_id=enterprise_customer_uuid,
            detail_resource='enrollments',
            should_traverse_pagination=True,
            querystring={'page_size': self.PAGE_SIZE},
        )


class EnterpriseDataV1ApiClient(EnterpriseDataApiClient):
    """
    Client for connecting to the Enterprise Data V1 API.
    """

    API_BASE_URL = urljoin(os.getenv('ANALYTICS_API_URL', default='') + '/', 'enterprise/api/v1')


class AnalyticsDataApiClient(EdxOAuth2APIClient):
    """
    Client for connecting to the Analytics Data API.
    """

    API_BASE_URL = urljoin(os.getenv('ANALYTICS_API_URL', default='') + '/', 'api/v0')
    PAGE_SIZE = os.getenv('PAGE_SIZE', default=1000)

    def get_enterprise_engagements(self, enterprise_customer_uuid):
        return self._load_data(
            'enterprise',
            resource_id=enterprise_customer_uuid,
            detail_resource='engagements',
            should_traverse_pagination=True,
            querystring={'page_size': self.PAGE_SIZE},
        )



"""
Module for interacting with the fact_enrollment_admin_dash table.
"""
from datetime import date, datetime
from typing import Optional, Tuple
from uuid import UUID

from enterprise_data.cache.decorators import cache_it

from ..filters import FactCompletionAdminDashFilters, FactEnrollmentAdminDashFilters
from ..queries import FactEnrollmentAdminDashQueries
from ..query_filters import QueryFilters
from ..utils import run_query
from .base import BaseTable


class FactEnrollmentAdminDashTable(BaseTable):
    """
    Class for communicating with the fact_enrollment_admin_dash table.
    """
    queries = FactEnrollmentAdminDashQueries()
    enrollment_filters = FactEnrollmentAdminDashFilters()
    completion_filters = FactCompletionAdminDashFilters()

    def __get_common_query_filters(
            self, enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date,
    ) -> (QueryFilters, dict):
        """
        Utility method to get query filters common in most usages below.

        This will return a tuple containing the query filters list and the dictionary of query parameters that
        will be used in the query.

        It will contain the following query filters.
            1. enterprise_customer_uuid filter to filter records for an enterprise customer.
            2. enrollment_date range filter to filter records by enrollment date.
            3. group_uuid filter to filter records for learners who belong to the given group.
        """
        query_filters = QueryFilters([
            self.enrollment_filters.enterprise_customer_uuid_filter('enterprise_customer_uuid'),
            self.enrollment_filters.date_range_filter('enterprise_enrollment_date', 'start_date', 'end_date'),
        ])
        params = {
            'enterprise_customer_uuid': enterprise_customer_uuid,
            'start_date': start_date,
            'end_date': end_date,
        }

        response = self.enrollment_filters.enterprise_user_query_filter(
            group_uuid, enterprise_customer_uuid
        )
        if response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        return query_filters, params

    def __get_common_query_filters_for_completion(
        self, enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
    ) -> Tuple[QueryFilters, dict]:
        """
        Utility method to get query filters common in most usages below.

        This will return a tuple containing the query filters list and the dictionary of query parameters that
        will be used in the query.

        It will contain the following query filters.
            1. enterprise_customer_uuid filter to filter records for an enterprise customer.
            2. enrollment_date range filter to filter records by enrollment date.
            3. group_uuid filter to filter records for learners who belong to the given group.
            4. has_passed filter to filter records for successful completions.
        """
        query_filters = QueryFilters([
            self.completion_filters.enterprise_customer_uuid_filter('enterprise_customer_uuid'),
            self.completion_filters.date_range_filter('passed_date', 'start_date', 'end_date'),
            self.completion_filters.has_passed_filter(),
        ])
        params = {
            'enterprise_customer_uuid': enterprise_customer_uuid,
            'start_date': start_date,
            'end_date': end_date,
        }

        response = self.completion_filters.enterprise_user_query_filter(
            group_uuid, enterprise_customer_uuid
        )
        if response is not None:
            enterprise_user_id_in_filter, enterprise_user_id_params = response
            query_filters.append(enterprise_user_id_in_filter)
            params.update(enterprise_user_id_params)

        return query_filters, params

    def get_top_enterprises(self, count=10):
        """
        Get the top enterprises by enrollments.

        Arguments:
            count (int): The number of enterprises to return.

        Returns:
            list<str>: A list of enterprise UUIDs.
        """
        result = run_query(
            query=self.queries.get_top_enterprises_query(count),
            as_dict=False,
        )
        return [row[0] for row in result]

    @cache_it()
    def get_enrollment_count(
            self,
            enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date,
    ) -> int:
        """
        Get the total number of enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            (int): The total number of enrollments.
        """
        query_filters, query_filter_params = self.__get_common_query_filters(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        results = run_query(
            query=self.queries.get_enrollment_count_query(query_filters),
            params=query_filter_params
        )
        if not results:
            return 0
        return int(results[0][0] or 0)

    @cache_it()
    def get_all_enrollments(
            self,
            enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date,
            limit: int,
            offset: int,
    ) -> list:
        """
        Get all enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.

        Returns:
            list<dict>: A list of dictionaries containing the enrollment data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        return run_query(
            query=self.queries.get_all_enrollments_query(query_filters),
            params={
                **query_filter_params,
                'limit': limit,
                'offset': offset,
            },
            as_dict=True,
        )

    @cache_it()
    def get_enrollment_date_range(self, enterprise_customer_uuid: UUID):
        """
        Get the enrollment date range for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID | str): The UUID of the enterprise customer.

        Returns:
            (tuple<date, date>): The minimum and maximum enrollment dates.
        """
        results = run_query(
            query=self.queries.get_enrollment_date_range_query(),
            params={'enterprise_customer_uuid': enterprise_customer_uuid}
        )
        if not results:
            return None, None
        min_date, max_date = results[0]

        # We should return date objects, not datetime objects
        # This is done to counter cases where database values are datetime objects.
        if min_date and isinstance(min_date, datetime):
            min_date = min_date.date()
        if max_date and isinstance(max_date, datetime):
            max_date = max_date.date()

        return min_date, max_date

    @cache_it()
    def get_enrollment_and_course_count(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the enrollment and course count for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            (tuple<int, int>): The enrollment and course count.
        """
        results = run_query(
            query=self.queries.get_enrollment_and_course_count_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            }
        )
        if not results:
            return 0, 0
        return tuple(results[0])

    @cache_it()
    def get_completion_count(
        self, enterprise_customer_uuid: UUID, group_uuid: Optional[UUID], start_date: date, end_date: date
    ):
        """
        Get the completion count for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            int: The completion count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_completion(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )
        results = run_query(
            query=self.queries.get_completion_count_query(query_filters),
            params=query_filter_params,
            as_dict=False,
        )
        # Handle empty results defensively
        if not results or not results[0] or len(results[0]) == 0:
            return 0

        return int(results[0][0] or 0)

    @cache_it()
    def get_top_courses_by_enrollments(
            self, enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date
    ) -> list:
        """
        Get the top courses enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the course key, course_title and enrollment count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        return run_query(
            query=self.queries.get_top_courses_by_enrollments_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_top_subjects_by_enrollments(
            self,
            enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date,
    ) -> list:
        """
        Get the top subjects by enrollments for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the subject and enrollment count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        return run_query(
            query=self.queries.get_top_subjects_by_enrollments_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_enrolment_time_series_data(
            self, enterprise_customer_uuid: UUID,
            group_uuid: Optional[UUID],
            start_date: date,
            end_date: date
    ) -> list:
        """
        Get the enrollment time series data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the date and enrollment count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )

        return run_query(
            query=self.queries.get_enrolment_time_series_data_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_all_completions(
        self,
        enterprise_customer_uuid: UUID,
        group_uuid: Optional[UUID],
        start_date: date,
        end_date: date,
        limit: int,
        offset: int,
    ):
        """
        Get all completions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            group_uuid (UUID): The UUID of the group.
            end_date (date): The end date.
            limit (int): The maximum number of records to return.
            offset (int): The number of records to skip.

        Returns:
            list<dict>: A list of dictionaries containing the completions data.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_completion(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )
        return run_query(
            query=self.queries.get_all_completions_query(query_filters),
            params={
                **query_filter_params,
                'limit': limit,
                'offset': offset,
            },
            as_dict=True,
        )

    @cache_it()
    def get_top_courses_by_completions(
        self, enterprise_customer_uuid: UUID, group_uuid: Optional[UUID], start_date: date, end_date: date
    ):
        """
        Get the top courses by completion for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the course key, course_title and completion count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_completion(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )
        return run_query(
            query=self.queries.get_top_courses_by_completions_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_top_subjects_by_completions(
        self, enterprise_customer_uuid: UUID, group_uuid: Optional[UUID], start_date: date, end_date: date
    ):
        """
        Get the top subjects by completions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the subject and completion count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_completion(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )
        return run_query(
            query=self.queries.get_top_subjects_by_completions_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )

    @cache_it()
    def get_completions_time_series_data(
        self, enterprise_customer_uuid: UUID, group_uuid: Optional[UUID], start_date: date, end_date: date
    ):
        """
        Get the completions time series data for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            group_uuid (UUID): The UUID of the group.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            list<dict>: A list of dictionaries containing the date and completion count.
        """
        query_filters, query_filter_params = self.__get_common_query_filters_for_completion(
            enterprise_customer_uuid, group_uuid, start_date, end_date
        )
        return run_query(
            query=self.queries.get_completions_time_series_data_query(query_filters),
            params=query_filter_params,
            as_dict=True,
        )



"""
Query filter for between operation.
"""
from .base import QueryFilter


class INQueryFilter(QueryFilter):
    """
    Query filter for IN operation.
    """

    def __init__(self, column: str, values: list = None, values_placeholders: list = None):
        """
        Initialize the filter.

        This will also validate arguments.
        """
        self.validate_argument_exclusivity(values, values_placeholders)

        self.column = column
        self.values = values
        self.values_placeholders = values_placeholders

    def to_sql(self) -> str:
        if self.values is not None:
            return f'{self.column} IN {str(tuple(self.value_to_sql(self.values)))}'
        else:
            return f'{self.column} IN {tuple(f"%({item})s" for item in self.values_placeholders)}'



"""
Views for enterprise admin api v1.
"""
from datetime import date, datetime

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import filters, viewsets
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND
from rest_framework.views import APIView

from enterprise_data.admin_analytics.data_loaders import fetch_max_enrollment_datetime
from enterprise_data.admin_analytics.database.tables import (
    FactEngagementAdminDashTable,
    FactEnrollmentAdminDashTable,
    SkillsDailyRollupAdminDashTable,
)
from enterprise_data.api.v1 import serializers
from enterprise_data.models import (
    EnterpriseAdminLearnerProgress,
    EnterpriseAdminSummarizeInsights,
    EnterpriseExecEdLCModulePerformance,
    EnterpriseGroupMembership,
    EnterpriseSubsidyBudget,
)
from enterprise_data.utils import timer

from .base import EnterpriseViewSetMixin


class EnterpriseAdminInsightsView(APIView):
    """
    API for getting the enterprise admin insights.
    """

    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin insights
        """
        response_data = {}
        learner_progress = {}
        learner_engagement = {}

        try:
            learner_progress = EnterpriseAdminLearnerProgress.objects.get(
                enterprise_customer_uuid=enterprise_id
            )
            learner_progress = serializers.EnterpriseAdminLearnerProgressSerializer(
                learner_progress
            ).data
            response_data["learner_progress"] = learner_progress
        except EnterpriseAdminLearnerProgress.DoesNotExist:
            pass

        try:
            learner_engagement = EnterpriseAdminSummarizeInsights.objects.get(
                enterprise_customer_uuid=enterprise_id
            )
            learner_engagement = serializers.EnterpriseAdminSummarizeInsightsSerializer(
                learner_engagement
            ).data
            response_data["learner_engagement"] = learner_engagement
        except EnterpriseAdminSummarizeInsights.DoesNotExist:
            pass

        status = HTTP_200_OK
        if learner_progress == {} and learner_engagement == {}:
            status = HTTP_404_NOT_FOUND

        return Response(data=response_data, status=status)


class EnterpriseAdminAnalyticsAggregatesView(APIView):
    """
    API for getting the enterprise admin analytics aggregates.
    """

    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin aggregate data.
        """
        # Validate the enterprise_id
        enterprise_id = enterprise_id.replace('-', '')
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(
            data=request.GET
        )
        serializer.is_valid(raise_exception=True)

        last_updated_at = fetch_max_enrollment_datetime()
        min_enrollment_date, max_enrollment_date = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_id,
        )

        start_date = serializer.data.get(
            'start_date', min_enrollment_date
        )
        end_date = serializer.data.get('end_date', datetime.today())
        group_uuid = serializer.data.get('group_uuid', None)

        enrolls, courses = FactEnrollmentAdminDashTable().get_enrollment_and_course_count(
            enterprise_id, start_date, end_date,
        )
        completions = FactEnrollmentAdminDashTable().get_completion_count(
            enterprise_id, group_uuid, start_date, end_date,
        )
        hours, sessions = FactEngagementAdminDashTable().get_learning_hours_and_daily_sessions(
            enterprise_id, start_date, end_date,
        )

        return Response(
            data={
                'enrolls': enrolls,
                'courses': courses,
                'completions': completions,
                'hours': hours,
                'sessions': sessions,
                'last_updated_at': last_updated_at if last_updated_at else None,
                'min_enrollment_date': min_enrollment_date,
                'max_enrollment_date': max_enrollment_date,
            },
            status=HTTP_200_OK,
        )


class EnterpriseAdminAnalyticsSkillsView(APIView):
    """
    API for getting the enterprise admin analytics skills data.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required(
        'can_access_enterprise', fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """HTTP GET endpoint to retrieve the enterprise admin skills aggregated data.

        Args:
            request (HttpRequest): request object
            enterprise_id (str): UUID of the enterprise customer

        Returns:
            response(HttpResponse): response object
        """
        enterprise_id = enterprise_id.replace('-', '')
        serializer = serializers.AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        if (start_date := serializer.data.get('start_date')) is None:
            start_date, _ = FactEnrollmentAdminDashTable().get_enrollment_date_range(enterprise_id)

        end_date = serializer.data.get('end_date', date.today())

        with timer('top_skills'):
            skills = SkillsDailyRollupAdminDashTable().get_top_skills(enterprise_id, start_date, end_date)

        with timer('top_skills_by_enrollments'):
            top_skills_by_enrollments = SkillsDailyRollupAdminDashTable().get_top_skills_by_enrollment(
                enterprise_id,
                start_date,
                end_date
            )
        with timer('top_skills_by_completions'):
            top_skills_by_completions = SkillsDailyRollupAdminDashTable().get_top_skills_by_completion(
                enterprise_id,
                start_date,
                end_date
            )

        response_data = {
            "top_skills": skills,
            "top_skills_by_enrollments": top_skills_by_enrollments,
            "top_skills_by_completions": top_skills_by_completions,
        }

        return Response(data=response_data, status=HTTP_200_OK)


class EnterpriseExecEdLCModulePerformanceViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    View to for getting enterprise exec ed learner module performance records.
    """
    serializer_class = serializers.EnterpriseExecEdLCModulePerformanceSerializer
    filter_backends = (filters.OrderingFilter, filters.SearchFilter)
    ordering_fields = '__all__'
    ordering = ['username', 'module_name', 'last_access']
    search_fields = (
        'username',
        'course_name'
    )

    def get_queryset(self):
        """
        Return the queryset of EnterpriseExecEdLCModulePerformance objects.
        """
        return EnterpriseExecEdLCModulePerformance.objects.filter(
            enterprise_customer_uuid=self.kwargs['enterprise_id'],
        )


class EnterpriseBudgetView(APIView):
    """
    View for getting budgets information for an enterprise.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required("can_access_enterprise", fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """
        Return the queryset of EnterpriseSubsidyBudget objects.
        """
        budgets = EnterpriseSubsidyBudget.objects.filter(
            enterprise_customer_uuid=enterprise_uuid,
        ).values(
            'subsidy_access_policy_uuid',
            'subsidy_access_policy_display_name',
        )

        serializer = serializers.EnterpriseBudgetSerializer(budgets, many=True)
        return Response(serializer.data)


class EnterpriseGroupMembershipView(APIView):
    """
    View for getting Group Memberships information for an enterprise.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required("can_access_enterprise", fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """
        Returns the groups and budgets for an enterprise.
        """
        groups = EnterpriseGroupMembership.objects.filter(
            enterprise_customer_id=enterprise_uuid,
            group_type='flex',
        ).values('enterprise_group_uuid', 'enterprise_group_name').distinct()

        serializer = serializers.EnterpriseGroupMembershipSerializer(groups, many=True)
        return Response(serializer.data)



# Generated by Django 3.2.23 on 2024-07-18 07:04

from django.db import migrations, models

import enterprise_data.utils


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0039_auto_20240212_1403'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='enterprise_group_name',
            field=models.CharField(max_length=25, null=True),
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='enterprise_group_uuid',
            field=models.UUIDField(null=True),
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='is_included_in_utilization',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='subscription_license_uuid',
            field=models.UUIDField(null=True),
        ),
        migrations.AlterField(
            model_name='enterpriselearnerenrollment',
            name='enterprise_enrollment_id',
            field=models.PositiveIntegerField(unique=True),
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='lpr_unique_id',
            field=models.PositiveBigIntegerField(default=enterprise_data.utils.get_unique_id, primary_key=True, serialize=False),
        ),
        migrations.AlterField(
            model_name='enterpriselearnerenrollment',
            name='enterprise_enrollment_id',
            field=models.PositiveIntegerField(null=True),
        ),
    ]



"""
Base Query Filter.
"""
from abc import ABC, abstractmethod


class QueryFilter(ABC):
    """
    Base Query filter class for functionality common to all filters.
    """
    @abstractmethod
    def to_sql(self) -> str:
        """
        Convert the filter to a SQL string.
        """
        raise NotImplementedError

    @staticmethod
    def validate_argument_exclusivity(value, value_placeholder):
        """
        Validate that that arguments are mutually exclusive.

        This method will make sure that at-least one of the value or value_placeholder is provided. it will also
        make sure that both value and value_placeholder are not provided at the same time.

        Arguments:
            value: The value of the filter.
            value_placeholder: The placeholder for the value.
        """
        if value is not None and value_placeholder is not None:
            raise ValueError('Both value and value_placeholder cannot be provided at the same time.')
        if value is None and value_placeholder is None:
            raise ValueError('Either value or value_placeholder must be provided.')
        return True

    def value_to_sql(self, value):
        """
        Convert the given value to a string in a way that it can be used inside WHERE clause of SQL query.

        Note: This method is meant to evolve according to user needs. currently it handles the following data types.
        More types can be added as needed.

        1. str: Encloses the string in single quotes.
        2. list<str>: Encloses the list of strings in a tuple after calling this method recursively on each item.

        Arguments:
            value (Any): The value to format.

        Returns:
            (Any): The formatted value.
        """
        if isinstance(value, str):
            return f"'{value}'"
        if isinstance(value, list):
            return [self.value_to_sql(item) for item in value]

        return value


class QueryFilters(list):
    """
    A list of QueryFilter objects.
    """
    def to_sql(self) -> str:
        """
        Convert the filters to a SQL string.
        """
        return ' AND '.join([_filter.to_sql() for _filter in self])
